#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <sys/stat.h>
#include <syslog.h>
#include <time.h>

#ifndef INADDR_NONE
# define INADDR_NONE ((in_addr_t)-1)
#endif

#include "rbldnsd.h"
#include "dns.h"
#include "mempool.h"

/* TODO:
 *
 * maybe multi-children environment.  Seems to be unnecessary.
 *
 * review all names to be consistent
 *
 * implement TCP mode (in separate process, to reduce number of syscalls)
 *
 * implement AXFR query - stupid idea but AXFR is widely used
 *
 * maybe: AXFR client to convert DNS zones into rbldnsd fomat
 *
 * control socket for reloads.  This may be done in a separate
 *   process which will listen for requests on control socket and
 *   send signal to it's child.  This will also give us an ability
 *   to monitor and restart failed child (ala supervise)
 *
 * understand IP address ranges in ip4[v]sets, e.g.
 *   4.5.16-4.5.19 (4.5.16.0-4.5.19.255)
 *
 * allow e.g. 217.23.134.193/29: 217.23.134.193 is NOT on the
 *   range boundary, correct form is 217.23.134.192/29.  Some people
 *   want any variant, but some prefer for strict control, so it
 *   will be easier to catch errors.  Run-time configurable?  Maybe
 *   per-dataset, or per-datafile?  Errm, how to specify that all? ;)
 *
 * allow 127.0.0.1/255.255.255.0 form?  Seems to be redundrand
 *   (CIDR exists), and it's not obvious how to handle masks like
 *   255.0.255.0 (what a mess)
 *
 * add `fnmatch' (or dnpat) type, to handle shell-style wildcards
 *   (no regexps please)
 *
 * how about AAAA queries?  IPv6-enabled sendmail on Solaris9 requests
 *   AAAA records (2.0.0.127.blocklist.example.com AAAA).  Currently,
 *   bind chokes on rbldnsd refusal here (returning SERVFAIL to sendmail),
 *   so the whole thing does not work.  Should rbldnsd make "fake"
 *   AAAA records from existing A records here?  Note that asking AAAA
 *   in this case is just a waste of resources.
 */

/* what a mess... this routine is to work around various snprintf
 * implementations.  It never return <1 or value greather than
 * size of buffer: i.e. it returns number of chars _actually written_
 * to a buffer.
 * Maybe replace this with an alternative (simplistic) implementation,
 * only %d/%u/%s, with additional %S to print untrusted data replacing
 * control chars with something sane, and to print `...' for arguments
 * that aren't fit (e.g. "%.5s", "1234567" will print `12...') ?
 */

static int vssprintf(char *buf, int bufsz, const char *fmt, va_list ap) {
  int r = vsnprintf(buf, bufsz, fmt, ap);
  return r < 0 ? 0 : r >= bufsz ? bufsz - 1 : r;
}
static int ssprintf(char *buf, int bufsz, const char *fmt, ...) {
  va_list ap;
  va_start(ap, fmt);
  bufsz = vssprintf(buf, bufsz, fmt, ap);
  va_end(ap);
  return bufsz;
}

static char *progname; /* limited to 32 chars */
static int logto;
#define LOGTO_STDOUT 0x01
#define LOGTO_SYSLOG 0x02

static void error(int errnum, const char *fmt, ...) {
  char buf[256];
  int l, pl;
  va_list ap;
  l = pl = ssprintf(buf, sizeof(buf), "%.30s: ", progname);
  va_start(ap, fmt);
  l += vssprintf(buf + l, sizeof(buf) - l, fmt, ap);
  if (errnum)
    l += ssprintf(buf + l, sizeof(buf) - l, ": %.50s", strerror(errnum));
  if (logto & LOGTO_SYSLOG) {
    fmt = buf + pl;
    syslog(LOG_ERR, strchr(fmt, '%') ? "%s" : fmt, fmt);
  }
  buf[l++] = '\n';
  write(2, buf, l);
  _exit(1);
}


/* Each zone is composed of a set of zonedatas.
 * There is a global list of zonedatas, each
 * with a timestamp etc.
 * Each zonedata is composed of a list of files.
 */

struct zonefile {
  time_t stamp;			/* last timestamp of this file */
  struct zonefile *next;	/* next file in list */
  const char *name;		/* name of this file */
};

struct zonedata {
  const struct zonetype *type;	/* type of this data */
  z_queryfn *queryfn;		/* cached from type */
  unsigned qfilter;		/* cached from type */
  struct zonedata *data;	/* type-specific data */
  time_t stamp;			/* timestamp: when loaded */
  const char *spec;		/* original specification */
  struct zonefile *file;	/* list of files for this data */
  struct zonedata *next;	/* next in global list */
};

struct zonedatalist {
  struct zonedata *data;
  struct zonedatalist *next;
};

struct zone {
  char *name;			/* name of the zone */
  int loaded;			/* true if loaded ok */
  unsigned char *dn;		/* domain name */
  unsigned dnlen;		/* length of dn */
  struct zonedatalist *dlist;	/* list of datas */
  struct zone *next;		/* next in list */
};

static unsigned ttl = 2048;	/* default record TTL */
static int recheck = 60;	/* interval between checks for reload */
static int verbose; /* = 0 */	/* log requests */
static int initialized;

/* a list of zonetypes. */
static const struct zonetype *zonetypes[] = {
  &ip4vset_zone,
  &ip4set_zone,
  &dnvset_zone,
  &dnset_zone,
  &generic_zone
};

static struct zonedata *zonedatas;

static struct {
  const struct zonedata *zd;
  const char *fname;
  int warns;
} curloading;

void
vzlog(int level, int lineno, const char *fmt, va_list ap) {
  char buf[1024];
  int pl, l;
  l = pl = (logto & LOGTO_STDOUT) ?
       ssprintf(buf, sizeof(buf), "%.30s: ", progname) : 0;
  if (curloading.zd) {
    l += ssprintf(buf + l, sizeof(buf) - l, "%s:%.50s: ",
		  curloading.zd->type->name, curloading.zd->spec);
    if (curloading.fname) {
      l += ssprintf(buf + l, sizeof(buf) - l, "%.60s",
		    curloading.fname);
      l += ssprintf(buf + l, sizeof(buf) - l,
		    lineno ? "(%d): " : ": ", lineno);
    }
  }
  l += vssprintf(buf + l, sizeof(buf) - l, fmt, ap);
  if (logto & LOGTO_SYSLOG) {
    fmt = buf + pl;
    syslog(level, strchr(fmt, '%') ? "%s" : fmt, fmt);
  }
  if (logto & LOGTO_STDOUT) {
    buf[l++] = '\n';
    write(1, buf, l);
  }
}

void zlog(int level, int lineno, const char *fmt, ...) {
  va_list ap;
  va_start(ap, fmt);
  vzlog(level, lineno, fmt, ap);
  va_end(ap);
}

#define MAXWARN 5

void zwarn(int lineno, const char *fmt, ...) {
  if (++curloading.warns <= MAXWARN) { /* prevent syslog flood */
    va_list ap;
    va_start(ap, fmt);
    zlog(LOG_WARNING, lineno, fmt, ap);
    va_end(ap);
  }
}

void zloaded(const char *fmt, ...) {
  char buf[128];
  va_list ap;
  struct tm *tm = gmtime(&curloading.zd->stamp);
  if (curloading.warns > MAXWARN)
    zlog(LOG_WARNING, 0, "%d more warnings suppressed",
         curloading.warns - MAXWARN);
  va_start(ap, fmt);
  vsnprintf(buf, sizeof(buf), fmt, ap);
  va_end(ap);
  zlog(LOG_INFO, 0, "%04d-%02d-%02d %02d:%02d:%02d: %s",
       tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday,
       tm->tm_hour, tm->tm_min, tm->tm_sec,
       buf);
}

static struct zonedata *newzonedata(char *spec) {
  /* type:file,file,file... */
  struct zonedata *z;
  char *f;
  struct zonefile **zfp, *zf;
  static const char *const delims = ",:";
  unsigned n;

  f = strchr(spec, ':');
  if (!f)
    error(0, "invalid zone data specification `%.60s'", spec);
  *f++ = '\0';

  for(z = zonedatas; z; z = z->next)
    if (strcmp(z->type->name, spec) == 0 && strcmp(z->spec, f) == 0)
      return z;

  z = (struct zonedata *)emalloc(sizeof(*z));
  memset(z, 0, sizeof(*z));
  z->next = zonedatas;
  zonedatas = z;
  z->spec = strdup(f);

  n = 0;
  while(strcmp(spec, zonetypes[n]->name))
    if (++n >= sizeof(zonetypes)/sizeof(zonetypes[0]))
      error(0, "unknown zone type `%.60s'", spec);
  z->type = zonetypes[n];
  z->queryfn = z->type->queryfn;
  z->qfilter = z->type->qfilter;

  for(zfp = &z->file, f = strtok(f, delims); f; f = strtok(NULL, delims)) {
    zf = (struct zonefile *)emalloc(sizeof(*zf));
    zf->stamp = 0;
    zf->name = estrdup(f);
    *zfp = zf;
    zfp = &zf->next;
  }
  *zfp = NULL;
  if (!z->file)
    error(0, "missing filenames for %s", spec);

  return z;
}

static struct zone *addzone(struct zone *zlist, const char *spec) {
  struct zone *z, **zp;
  struct zonedatalist *zd, **zdp;
  char *p;
  char name[DNS_MAXDOMAIN+1];
  unsigned char dn[DNS_MAXDN];
  unsigned dnlen;

  p = strchr(spec, ':');
  if (!p || p - spec > DNS_MAXDOMAIN)
    error(0, "invalid zone spec `%.60s'", spec);

  memcpy(name, spec, p - spec);
  name[p - spec] = '\0';

  dnlen = dns_ptodn(name, dn, sizeof(dn));
  if (!dnlen)
    error(0, "invalid domain name `%s'", name);
  dns_dntol(dn, dn);

  zp = &zlist;
  for (;;) {
    if (!(z = *zp)) {
      *zp = z = (struct zone *)emalloc(sizeof(*z));
      memset(z, 0, sizeof(*z));
      z->dn = (unsigned char *)emalloc(dnlen);
      memcpy(z->dn, dn, dnlen);
      z->dnlen = dnlen;
      dns_dntop(dn, name, sizeof(name));
      z->name = estrdup(name);
      strcpy(z->name, name);
      break;
    }
    else if (z->dnlen == dnlen && memcmp(z->dn, dn, dnlen) == 0)
      break;
    else
      zp = &z->next;
  }

  zdp = &z->dlist;
  while(*zdp)
    zdp = &(*zdp)->next;

  zd = *zdp = emalloc(sizeof(*zd));
  zd->next = NULL;
  p = strdup(p+1);
  zd->data = newzonedata(p);
  free(p);

  return zlist;
}

static int loadzonedata(struct zonedata *z) {
  struct zonefile *zf;
  time_t stamp = 0;
  FILE *f;

  if (z->data)
    z->type->freefn(z->data);

  z->data = z->type->allocfn();
  if (!z->data)
    return 0;
  for(zf = z->file; zf; zf = zf->next) {
    curloading.fname = zf->name;
    f = fopen(zf->name, "r");
    if (!f) {
      zlog(LOG_ERR, 0, "unable to open file: %s", strerror(errno));
      return 0;
    }
    if (!z->type->loadfn(z->data, f)) {
      fclose(f);
      return 0;
    }
    if (ferror(f)) {
      zlog(LOG_ERR, 0, "error reading file `%.60s': %s",
           zf->name, strerror(errno));
      fclose(f);
      return 0;
    }
    fclose(f);
    if (zf->stamp > stamp)
      stamp = zf->stamp;
  }
  curloading.fname = NULL;
  z->stamp = stamp;

  if (!z->type->finishfn(z->data))
    return 0;

  return 1;
}

static void reload(struct zone *zl) {
  struct zonedata *zd;
  struct zonefile *zf;
  struct zonedatalist *zdl;
  int reloaded = 0;

  for(zd = zonedatas; zd; zd = zd->next) {
    int load = 0;

    memset(&curloading, 0, sizeof(curloading));
    curloading.zd = zd;

    for(zf = zd->file; zf; zf = zf->next) {
      struct stat st;
      if (stat(zf->name, &st) < 0) {
        zlog(LOG_ERR, 0, "unable to stat file `%.60s': %s",
	     zf->name, strerror(errno));
	load = -1;
	break;
      }
      else if (zf->stamp != st.st_mtime) {
        load = 1;
        zf->stamp = st.st_mtime;
      }
    }

    if (!load)
      continue;

    ++reloaded;

    if (load < 0 || !loadzonedata(zd)) {
      if (!initialized)
	error(0, "aborting");
      for (zf = zd->file; zf; zf = zf->next)
	zf->stamp = 0;
      if (zd->data) {
	zd->type->freefn(zd->data);
	zd->data = NULL;
      }
      zd->stamp = 0;
    }

  }

  curloading.zd = NULL;

  if (!reloaded)
    return;

  while(zl) {
    for(zdl = zl->dlist; zdl; zdl = zdl->next)
      if (!zdl->data->stamp)
	break;
    if (zdl) {
      zlog(LOG_WARNING, 0, "partially loaded zone %.60s will not be serviced",
	   zl->name);
      zl->loaded = 0;
    }
    else
      zl->loaded = 1;
    zl = zl->next;
  }
}

static int satoi(const char *s) {
  int n = 0;
  if (*s < '0' || *s > '9') return -1;
  do n = n * 10 + (*s++ - '0');
  while (*s >= '0' && *s <= '9');
  return *s ? -1 : n;
}

static void usage() {
   int i;
   printf(
"%s: rbl dns daemon version " VERSION "\n"
"Usage is: %s [options] zone...\n"
"where options are:\n"
" -u user[:group] - run as this user:group (rbldns)\n"
" -r rootdir - chroot to this directory\n"
" -w workdir - working directory with zone files\n"
" -b [address][:port] - bind to (listen on) this address (*:53)\n"
" -t ttl - TTL value set in answers (2048)\n"
" -c check - check for file updates every `check' secs (60)\n"
" -p pidfile - write backgrounded pid to specified file\n"
" -n - do not become a daemon\n"
" -v - verbose, log queries and answers\n"
"each zone specified using `name:type:file,file...'\n"
"syntax, repeated names constitute the same zone.\n"
"Available zone types:\n"
, progname, progname);
   for(i = 0; zonetypes[i]; ++i)
     printf(" %s %s\n", zonetypes[i]->name, zonetypes[i]->descr);
   exit(0);
}

static int init(int argc, char **argv, struct zone **zonep) {
  int c;
  char *p;
  char *user = "rbldns", *bindaddr = "";
  char *root = NULL, *workdir = NULL, *pidfile = NULL;
  FILE *fpid = NULL;
  uid_t uid = 0;
  gid_t gid = 0;
  struct sockaddr_in sin;
  int fd;
  int nodaemon = 0;

  if ((progname = strrchr(argv[0], '/')) != NULL)
    argv[0] = ++progname;
  else
    progname = argv[0];

  if (argc <= 1) usage();

  while((c = getopt(argc, argv, "u:r:b:w:t:c:p:nvh")) != EOF)
    switch(c) {
    case 'u': user = optarg; break;
    case 'r': root = optarg; break;
    case 'b': bindaddr = optarg; break;
    case 'w': workdir = optarg; break;
    case 'p': pidfile = optarg; break;
    case 't':
      if ((c = satoi(optarg)) < 0)
        error(0, "invalid ttl (-t) value `%.50s'", optarg);
      ttl = c;
      break;
    case 'c':
      if ((c = satoi(optarg)) < 0)
        error(0, "invalid check interval (-c) value `%.50s'", optarg);
      recheck = c;
      break;
    case 'n': nodaemon = 1; break;
    case 'v': ++verbose; break;
    case 'h': usage(); break;
    default: error(0, "type `%.50s -h' for help", progname);
    }

  if (!(argc -= optind))
    error(0, "no zone(s) to service specified (-h for help)");
  argv += optind;

  if (nodaemon)
    logto = LOGTO_STDOUT;
  else {
    tzset();
    openlog(progname, LOG_PID|LOG_NDELAY, LOG_DAEMON);
    logto = LOGTO_STDOUT | LOGTO_SYSLOG;
  }

  fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (fd < 0)
    error(errno, "unable to create listening socket");
  memset(&sin, 0, sizeof(sin));
  sin.sin_family = AF_INET;
  if (bindaddr) {
    if ((p = strchr(bindaddr, ':')) != NULL)
      *p++ = '\0';
    if (*bindaddr &&
        (sin.sin_addr.s_addr = inet_addr(bindaddr)) == INADDR_NONE) {
      struct hostent *he = gethostbyname(bindaddr);
      if (!he)
        error(0, "invalid bind address specified: `%.50s'", bindaddr);
      if (he->h_addrtype != AF_INET || he->h_length != sizeof(sin.sin_addr))
        error(0, "unexpected type of listening address `%.50s'", bindaddr);
      memcpy(&sin.sin_addr, he->h_addr_list[0], sizeof(sin.sin_addr));
      endhostent();
    }
  }
  else
    p = NULL;
  if (p && *p) {
    if ((c = satoi(p)) > 0)
      ;
    else {
      struct servent *se = getservbyname(p, "udp");
      if (!se)
        error(0, "%.50s/udp: unknown service", p);
      c = se->s_port;
      endservent();
    }
    p[-1] = ':';
  }
  else
    c = DNS_PORT;
  sin.sin_port = htons(c);
  if (bind(fd, (struct sockaddr *)&sin, sizeof(sin)) < 0)
    error(errno, "unable to bind to %.50s:%d", inet_ntoa(sin.sin_addr), c);

  c = 65536;
  do
    if (setsockopt(fd, SOL_SOCKET, SO_RCVBUF, &c, sizeof c) == 0)
      break;
  while ((c -= (c >> 5)) >= 1024);

  if ((p = strchr(user, ':')) != NULL)
    *p++ = '\0';
  if ((c = satoi(user)) >= 0)
    uid = c, gid = c;
  else {
    struct passwd *pw = getpwnam(user);
    if (!pw)
      error(0, "invalid user `%.50s' specified", user);
    uid = pw->pw_uid;
    gid = pw->pw_gid;
    endpwent();
  }
  if (!uid)
    error(0, "daemon should not run as root, specify -u option");
  if (p) {
    if ((c = satoi(p)) >= 0)
      gid = c;
    else {
      struct group *gr = getgrnam(p);
      if (!gr)
        error(0, "invalid group `%.50s' specified", p);
      gid = gr->gr_gid;
      endgrent();
    }
    p[-1] = ':';
  }

  if (pidfile && (fpid = fopen(pidfile, "w")) == NULL)
    error(errno, "unable to write pidfile");

  if (root && (chdir(root) < 0 || chroot(root) < 0))
    error(errno, "unable to chroot");
  if (workdir && chdir(workdir) < 0)
    error(errno, "unable to chdir to %.50s", workdir);

  if (setgroups(1, &gid) < 0 || setgid(gid) < 0 || setuid(uid) < 0)
    error(errno, "unable to setuid(%d:%d)", uid, gid);

  *zonep = NULL;
  for(c = 0; c < argc; ++c)
    *zonep = addzone(*zonep, argv[c]);

  reload(*zonep);
  initialized = 1;
  zlog(LOG_INFO, 0, "started");

  if (!nodaemon) {
    if (fork() > 0) exit(0);
    close(0); close(1); close(2);
    setsid();
    logto = LOGTO_SYSLOG;
  }
  if (fpid) {
    fprintf(fpid, "%lu\n", (unsigned long)getpid());
    fclose(fpid);
  }

  return fd;
}

/* DNS packet:
 * 0:1   identifier (client supplied)
 * 2     flags1
 *  0x80 QR response flag
 *  0x78 opcode: 0 query
 *  0x04 AA auth answer
 *  0x02 TC truncation flag
 *  0x01 RD recursion desired (may be set in query)
 * 3     flags2
 *  0x80 RA recursion available
 *  0x70 Z reserved
 *  0x0f RCODE:
 *     0 ok, 1 format error, 2 servfail, 3 nxdomain, 4 notimpl, 5 refused
 * 4:5   qdcount (numqueries)
 * 6:7   ancount (numanswers)
 * 8:9   nscount (numauthority)
 * 10:11 arcount (numadditional)
 */

struct dnspacket {
  unsigned char *c; /* current pointer */
  unsigned char *sans; /* start of answers */
  unsigned nans; /* number of answers */
  unsigned char *p; /* packet, where it begin */
  unsigned char *e; /* pointer to end-of-packet */
};

static unsigned process(struct dnspacket *p, const struct zone *zone) {
  unsigned char *q, *x;
  unsigned qlen, qcls, qtyp;
  unsigned char query[DNS_MAXDN+1];
  int nm, nf;

  q = p->p + 12;
  x = p->c;
  if (q >= x) return 0; /* short packet */
  if (p->p[2] & 0x80) return 0; /* response?! */
  if (p->p[4] || p->p[5] != 1) return 0; /* qdcount != 1 */
  /* check query's DN */
  x -= 4;
  while(*q)
    if (q + *q >= x) return 0;
    else q += *q + 1;
  qlen = ++q - (p->p + 12);
  dns_dntol(p->p + 12, query);
  qtyp = ((unsigned)(q[0]) << 8) | q[1];
  qcls = ((unsigned)(q[2]) << 8) | q[3];
  q += 4;

  /* from now on, we see (almost?) valid dns query, should reply */
  p->c = p->sans = q;
  p->nans = 0;

#define refuse(p,code) (p->p[2] = 0x84, p->p[3] = (code), (p->sans - p->p))
#define nxdomain(p) (p->p[2] = 0x84, p->p[3] = 3, (p->sans - p->p))

  /* construct reply packet */
  /* identifier already in place */
  /* flags will be set up later */
  /* p[4:5] (qdcount) already set up in query */
  p->p[6] = p->p[7] = 0; /* ancount */
  p->p[8] = p->p[9] = 0; /* nscount */
  p->p[10] = p->p[11] = 0; /* arcount */

  if (qcls != DNS_C_IN && qcls != DNS_C_ANY)
    return refuse(p,1); /* format error */
  if (p->p[2] & 126)
    return refuse(p,4); /* not implemented */
  switch(qtyp) {
  case DNS_T_ANY: qtyp = NSQUERY_ANY; break;
  case DNS_T_A:   qtyp = NSQUERY_A; break;
  case DNS_T_TXT: qtyp = NSQUERY_TXT; break;
  case DNS_T_NS:  qtyp = NSQUERY_NS; break;
  case DNS_T_SOA: qtyp = NSQUERY_SOA; break;
  default: return refuse(p,5); /* refused */
  }

  p->p[2] = 0x80; /* 0x81?! */
  if (qcls == DNS_C_IN) p->p[2] |= 0x04; /* AA */
  p->p[3] = 0;

  nm = nf = 0;

  for(; zone; zone = zone->next) {
    const struct zonedatalist *zdl;

    if (!zone->loaded) continue;
    if (zone->dnlen > qlen) continue;
    x = query + qlen - zone->dnlen;
    if (memcmp(zone->dn, x, zone->dnlen)) continue;
    q = query;
    while(q < x) q += *q + 1;
    if (q != x) continue;
    *q = 0;

    for(zdl = zone->dlist; zdl; zdl = zdl->next)
      if (zdl->data->qfilter & qtyp) {
	nm = 1;
	if (zdl->data && zdl->data->queryfn(zdl->data->data, p, query, qtyp))
	  nf = 1;
      }

    *q = zone->dn[0];

  }

  if (nf) {
    p->p[6] = p->nans >> 8; p->p[7] = p->nans;
    return p->c - p->p;
  }
  else if (nm)
    return nxdomain(p);
  else
    return refuse(p, 5); /* refused */

}

static int aexists(const struct dnspacket *p, unsigned typ,
                   const void *val, unsigned vlen) {
  const unsigned char *c, *e;
  for(c = p->sans, e = p->c; c < e; c = c + 12 + c[11]) {
    if (c[2] == (typ>>8) && c[3] == (typ&255) &&
        c[11] == vlen && memcmp(c + 12, val, vlen) == 0)
      return 1;
  }
  return 0;
}

int addrec_any(struct dnspacket *p, unsigned dtp,
               const void *data, unsigned dsz) {
  if (p->c + 12 + dsz >= p->e) return 0;
  if (aexists(p, dtp, data, dsz)) return 0;
  p->nans++;
  *p->c++ = 192; *p->c++ = 12; /* jump after header */
  *p->c++ = dtp >> 8; *p->c++ = dtp; /* dtype */
  *p->c++ = DNS_C_IN>>8; *p->c++ = DNS_C_IN; /* class */
  *p->c++ = ttl>>24; *p->c++ = ttl>>16; *p->c++ = ttl>>8; *p->c++ = ttl;
  *p->c++ = dsz>>8; *p->c++ = dsz; /* dsize */
  memcpy(p->c, data, dsz);
  p->c += dsz;
  return 1;
}

int
addrec_a(struct dnspacket *p, ip4addr_t aip) {
  aip = htonl(aip);
  return addrec_any(p, DNS_T_A, &aip, 4);
}

#if 0
  if (p->c + 16 >= p->e)
    return 0;
  aip = htonl(aip);
  if (aexists(p, DNS_C_IN, &aip, 4)) return 0;
  p->nans++;
  *p->c++ = 192; *p->c++ = 12;  /* jump after header */
  *p->c++ = DNS_T_A>>8; *p->c++ = DNS_T_A; /* type */
  *p->c++ = DNS_C_IN>>8; *p->c++ = DNS_C_IN; /* class */
  *p->c++ = ttl>>24; *p->c++ = ttl>>16; *p->c++ = ttl>>8; *p->c++ = ttl;
  *p->c++ = 0; *p->c++ = 4;
  memcpy(p->c, &aip, 4);
  p->c += 4;
  return 1;
}
#endif

int
addrec_txt(struct dnspacket *p, const char *txt, const char *subst) {
  unsigned sl;
  char sb[256];
  char *lp, *s, *const e = sb + 254;
  if (!txt) return 0;
  if (p->c + 13 >= p->e) return 0;
  lp = sb + 1;
  if (!subst) subst = "$";
  while(lp < e) {
    if ((s = strchr(txt, '$')) == NULL)
      s = (char*)txt + strlen(txt);
    sl = s - txt;
    if (lp + sl > e)
      sl = e - lp;
    memcpy(lp, txt, sl);
    lp += sl;
    if (!*s) break;
    sl = strlen(subst);
    if (lp + sl > e)
      sl = e - lp;
    memcpy(lp, subst, sl);
    lp += sl;
    txt = s + 1;
  }
  sl = lp - sb;
  sb[0] = sl - 1;
#if 1
  return addrec_any(p, DNS_T_TXT, sb, sl);
#else
  if (p->c + 12 + sl >= p->e) return 0;
  if (aexists(p, DNS_T_TXT, sb, sl)) return 0;
  p->nans++;
  *p->c++ = 192; *p->c++ = 12; /* jump to after header */
  *p->c++ = DNS_T_TXT>>8; *p->c++ = DNS_T_TXT;
  *p->c++ = DNS_C_IN>>8; *p->c++ = DNS_C_IN;
  *p->c++ = ttl>>24; *p->c++ = ttl>>16; *p->c++ = ttl>>8; *p->c++ = ttl;
  *p->c++ = 0; *p->c++ = sl; /* length of record */
  memcpy(p->c, sb, sl);
  p->c += sl;
  return 1;
#endif
}

static volatile int want_reload;
static void sighandler(int sig) {
  switch(sig) {
  case SIGALRM:
    alarm(recheck);
    want_reload = 1;
    break;
  case SIGHUP:
    want_reload = 1;
    break;
  case SIGTERM:
  case SIGINT:
    curloading.zd = NULL;
    zlog(LOG_INFO, 0, "terminating");
    exit(0);
  }
}

static sigset_t sigset_block; /* signals to block during zone reload */

static void setup_signals() {
  struct sigaction sa;
  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = sighandler;
  sigemptyset(&sigset_block);
  sigaction(SIGHUP, &sa, NULL);
  sigaddset(&sigset_block, SIGHUP);
  sigaction(SIGALRM, &sa, NULL);
  sigaddset(&sigset_block, SIGALRM);
  sigaction(SIGTERM, &sa, NULL);
  sigaction(SIGINT, &sa, NULL);
}

/* this routine should log to a file instead of syslog */

static void loganswer(struct dnspacket *pkt, struct sockaddr_in *sin) {
  char logbuf[DNS_MAXPACKET*3];
  char *s;
  unsigned char *p = pkt->p + 12;
  unsigned q;
  char *v;

  s = logbuf + sprintf(logbuf, "%s ", inet_ntoa(sin->sin_addr));
#ifdef MJT
/* yes, there should be an ability to suppress some logging! */
  if (memcmp(logbuf, "217.23.134.", 11) == 0 ||
      memcmp(logbuf, "127.", 4) == 0 ||
      memcmp(logbuf, "192.168.", 6) == 0
      ) return;
#endif
  s += dns_dntop(p, s, logbuf + sizeof(logbuf) - s - 100);
  *s++ = ' ';
  p += dns_dnlen(p);

  q = ((unsigned)p[0]<<8)|p[1];
  switch(q) {
  case DNS_T_A:   v = "A"; break;
  case DNS_T_TXT: v = "TXT"; break;
  case DNS_T_NS:  v = "NS"; break;
  case DNS_T_SOA: v = "SOA"; break;
  case DNS_T_MX:  v = "MX"; break;
  case DNS_T_ANY: v = "ANY"; break;
  default: s += sprintf(s, "type0x%x", q); v = NULL;
  }
  if (v) s += sprintf(s, "%s", v);
  *s++ = ' ';

  q = ((unsigned)p[2]<<8)|p[3];
  switch(q) {
  case DNS_C_IN: v = "IN"; break;
  case DNS_C_ANY: v = "ANY"; break;
  default: s += sprintf(s, "cls0x%x", q); v = NULL;
  }
  if (v) s += sprintf(s, "%s", v);
  *s++ = ':';
  *s++ = ' ';

  p = pkt->p;
  q = p[3];
  switch(q) {
  case 0: v = "NOERROR"; break;
  case 1: v = "FORMERR"; break;
  case 2: v = "SERVFAIL"; break;
  case 3: v = "NXDOMAIN"; break;
  case 4: v = "NOTIMPL"; break;
  case 5: v = "REFUSED"; break;
  default: s += sprintf(s, "code%u", q); v = NULL;
  }
  if (v) s += sprintf(s, "%s", v);
  s += sprintf(s, "/%u/%d", pkt->nans, pkt->c - pkt->p);
  if (logto & LOGTO_SYSLOG)
    syslog(LOG_INFO, strchr(logbuf, '%') ? "%s" : logbuf, logbuf);
  if (logto & LOGTO_STDOUT) {
    *s++ = '\n';
    write(1, logbuf, s - logbuf);
  }
}


int main(int argc, char **argv) {
  int fd;
  struct zone *zonelist;
  struct dnspacket p;
  unsigned char buf[DNS_MAXPACKET+1];
  fd = init(argc, argv, &zonelist);
  setup_signals();
  alarm(recheck);
  p.p = buf;
  p.e = buf + sizeof(buf);
  for(;;) {
    struct sockaddr_in sin;
    socklen_t sinl = sizeof(sin);
    int r;
    if (want_reload) {
      reload(zonelist);
      want_reload = 0;
    }
    r = recvfrom(fd, buf, sizeof(buf), 0, (struct sockaddr *)&sin, &sinl);
    if (r < 0)
      continue;
    p.c = buf + r;
    if (!(r = process(&p, zonelist))) {
      continue;
    }
    if (verbose)
      loganswer(&p, &sin);
    while(sendto(fd, buf, r, 0, (struct sockaddr *)&sin, sizeof(sin)) < 0)
      if (errno != EINTR) break;
  }
  return 0;
}

void *emalloc(unsigned size) {
  void *ptr = malloc(size);
  if (!ptr)
    oom();
  return ptr;
}

void *erealloc(void *ptr, unsigned size) {
  void *nptr = realloc(ptr, size);
  if (!nptr)
    oom();
  return nptr;
}

char *estrdup(const char *str) {
  char *s = strdup(str);
  if (!s) oom();
  return s;
}

void *mp_ealloc(struct mempool *mp, unsigned size) {
  void *p = mp_alloc(mp, size);
  if (!p) oom();
  return p;
}

char *mp_estrdup(struct mempool *mp, const char *str) {
  str = mp_strdup(mp, str);
  if (!str) oom();
  return (char*)str;
}

const char *mp_edstrdup(struct mempool *mp, const char *str) {
  str = mp_dstrdup(mp, str);
  if (!str) oom();
  return str;
}

void oom() {
  if (initialized)
    zlog(LOG_ERR, 0, "out of memory loading zone (zone will be empty)");
  else
    error(0, "out of memory");
}
